import setuptools
import sys

assert sys.version_info >= (3, 8)

setuptools.setup(
  name='op-service',
  version='0.1.6',
  author='Alex Exarhos',
  author_email='AlexExarhos@op-dev.io',
  packages=[
    'op_service'
  ],
  install_requires=[
    'iso8601 >= 0.1.11',
    'op-alarm >= 1.0.0',
    'op-service-client >= 0.1.1',
    'op-storage >= 1.0.0',
    'PyJWT >= 2.8.0',
    'python-multipart >= 0.0.6',
    'starlette >= 0.31.1'
  ]
)
